'use strict'

import {
  ServerConnection
} from '@jupyterlab/services';

import {
  VDomModel
} from '@jupyterlab/apputils';

import * as semver from 'semver';

import {
  Searcher, ISearchResult
} from './query';

/**
 * Information about an extension
 */
export
interface IEntry {
  name: string;
  description: string;
  python_package: string | null;
  installed: boolean;
  enabled: boolean;
  status: 'ok' | 'warning' | 'error' | 'deprecated' | null;
  latest_version: string;
  installed_version: string;
}


/**
 * Information about an installed extension
 */
export
interface IInstalledEntry {
  name: string;
  description: string;
  enabled: boolean;
  core: boolean;
  latest_version: string;
  installed_version: string;
  status: 'ok' | 'warning' | 'error' | 'deprecated' | null;
}


/**
 * The server API path for querying/modifying installed extensions.
 */
const EXTENSION_API_PATH = "lab/api/extensions"

/**
 * Extension actions that the server API accepts
 */
export
type Action = 'install' | 'uninstall' | 'enable' | 'disable';


/**
 * Model for an extension list.
 */
export
class ListModel extends VDomModel {
  constructor() {
    super();
    this._installed = [];
    this._searchResult = [];
    this.serverConnectionSettings = ServerConnection.makeSettings({withCredentials: true});
  }

  /**
   * Utility function to check whether an entry can be updated.
   *
   * @param entry The entry to check.
   */
  static entryHasUpdate(entry: IEntry): boolean {
    if (!entry.installed || !entry.latest_version) {
      return false;
    }
    return semver.lt(entry.installed_version, entry.latest_version);
  }

  /**
   * A readonly array of the installed extensions.
   */
  get installed(): ReadonlyArray<IEntry> {
    return this._installed;
  }

  /**
   * A readonly array containing the latest search result
   */
  get searchResult(): ReadonlyArray<IEntry> {
    return this._searchResult;
  }

  /**
   * Translate search results from an npm repository query into entries.
   *
   * @param res Promise to an npm query result.
   */
  protected async translateSearchResult(res: Promise<ISearchResult>): Promise<{[key: string]: IEntry}> {
    let entries: {[key: string]: IEntry} = {};
    for (let obj of (await res).objects) {
      let pkg = obj.package;
      entries[pkg.name] = {
        name: pkg.name,
        description: pkg.description,
        python_package: null,
        installed: false,
        enabled: false,
        status: null,
        latest_version: pkg.version,
        installed_version: '',
      };
    }
    return entries;
  }

  /**
   * Translate installed extensions information from the server into entries.
   *
   * @param res Promise to the server reply data.
   */
  protected async translateInstalled(res: Promise<IInstalledEntry[]>): Promise<{[key: string]: IEntry}> {
    const promises = [];
    const entries: {[key: string]: IEntry} = {};
    for (let pkg of await res) {
      promises.push(res.then((info) => {
        entries[pkg.name] = {
          name: pkg.name,
          description: pkg.description,
          python_package: null,
          installed: true,
          enabled: pkg.enabled,
          status: pkg.status,
          latest_version: pkg.latest_version,
          installed_version: pkg.installed_version,
        };
      }));
    }
    return Promise.all(promises).then(() => {
      return entries;
    });
  }

  /**
   * The current NPM repository search query.
   * 
   * Setting its value triggers a new search.
   */
  get query(): string {
    return this._query
  }
  set query(value: string) {
    this._query = value;
    this.update();
  }

  /**
   * The current NPM repository search page.
   *
   * The npm repository search is paginated by the `pagination` attribute.
   * The `page` value selects which page is used.
   *
   * Setting its value triggers a new search.
   */
  get page(): number {
    return this._page
  }
  set page(value: number) {
    this._page = value;
    this.update();
  }

  /**
   * The NPM repository search pagination.
   *
   * The npm repository search is paginated by the `pagination` attribute.
   * The `page` value selects which page is used.
   *
   * Setting its value triggers a new search.
   */
  get pagination(): number {
    return this._pagination
  }
  set pagination(value: number) {
    this._pagination = value;
    this.update();
  }

  /**
   * The total number of results in the current search.
   */
  get totalEntries(): number {
    return this._totalEntries;
  }

  /**
   * Make a request to the server for info about its installed extensions.
   */
  protected fetchInstalled(): Promise<IInstalledEntry[]> {
    let request: ServerConnection.IRequest = {
      url: EXTENSION_API_PATH,
    };
    return ServerConnection.makeRequest(request, this.serverConnectionSettings).then((response) => {
      return response.data as IInstalledEntry[];
    });
  }

  /**
   * Initialize the model.
   */
  initialize() {
    this.update().then(() => {
      this.initialized = true;
      this.stateChanged.emit(undefined);
    }).catch(() => {
      this.initialized = true;
      this.stateChanged.emit(undefined);
    });
  }

  /**
   * Update the current model.
   *
   * This will query the NPM repository, and the notebook server.
   *
   * Emits the `stateChanged` signal on succesfull completion.
   */
  protected async update() {
    let search = this.searcher.searchExtensions(this.query, this.page, this.pagination);
    let searchMapPromise = this.translateSearchResult(search);
    let installedMap = await this.translateInstalled(this.fetchInstalled());
    let searchMap;
    try {
      searchMap = await searchMapPromise;
      this.offline = false;
      this.errorMessage = undefined;
    } catch (reason) {
      searchMap = {};
      this.offline = true;
      this.errorMessage = reason.toString();
    }
    let installed: IEntry[] = [];
    for (let key of Object.keys(installedMap)) {
      installed.push(installedMap[key]);
    }
    this._installed = installed;

    let searchResult: IEntry[] = [];
    for (let key of Object.keys(searchMap)) {
      if (installedMap[key] === undefined) {
        searchResult.push(searchMap[key]);
      } else {
        searchResult.push(installedMap[key]);
      }
    }
    this._searchResult = searchResult;
    try {
      this._totalEntries = (await search).total;
    } catch (error) {
      this._totalEntries = 0;
    }
    this.stateChanged.emit(undefined);
  }

  /**
   * Send a request to the server to perform an action on an extension.
   *
   * @param action A valid action to perform.
   * @param entry The extension to perform the action on.
   */
  protected _performAction(action: string, entry: IEntry) {
    let request: ServerConnection.IRequest = {
      url: EXTENSION_API_PATH,
      method: 'POST',
      data: JSON.stringify({
        cmd: action,
        extension_name: entry.name,
      }),
    };
    return ServerConnection.makeRequest(request, this.serverConnectionSettings).then((response) => {
      return response.data as IInstalledEntry[];
    });
  }

  /**
   * Install an extension.
   *
   * @param entry An entry indicating which extension to install.
   */
  install(entry: IEntry) {
    if (entry.installed) {
      throw new Error(`Already installed: ${entry.name}`);
    }
    this._performAction('install', entry).then((data) => {
      this.update();
    });
  }

  /**
   * Uninstall an extension.
   *
   * @param entry An entry indicating which extension to uninstall.
   */
  uninstall(entry: IEntry) {
    if (!entry.installed) {
      throw new Error(`Not installed, cannot uninstall: ${entry.name}`);
    }
    this._performAction('uninstall', entry).then((data) => {
      this.update();
    });;
  }

  /**
   * Enable an extension.
   *
   * @param entry An entry indicating which extension to enable.
   */
  enable(entry: IEntry) {
    if (entry.enabled) {
      throw new Error(`Already enabled: ${entry.name}`);
    }
    this._performAction('enable', entry).then((data) => {
      this.update();
    });;
  }

  /**
   * Disable an extension.
   *
   * @param entry An entry indicating which extension to disable.
   */
  disable(entry: IEntry) {
    if (!entry.enabled) {
      throw new Error(`Already disabled: ${entry.name}`);
    }
    this._performAction('disable', entry).then((data) => {
      this.update();
    });;
  }

  /**
   * Set to true if an error occurs when trying to query the NPM repository.
   */
  offline: boolean | undefined;

  /**
   * Contains an error message if an error occurred when updating the model.
   */
  errorMessage: string | undefined;

  /**
   * Whether the model has finished async initialization.
   */
  initialized: boolean = false;

  private _query: string = '';
  private _page: number = 0;
  private _pagination: number = 250;
  private _totalEntries: number = 0;

  protected _installed: IEntry[];
  protected _searchResult: IEntry[];

  /**
   * Settings for connecting to the notebook server.
   */
  protected serverConnectionSettings: ServerConnection.ISettings;

  /**
   * A helper for performing searches of jupyterlab extensions on the NPM repository.
   */
  protected searcher = new Searcher();
}
