import { ServerConnection } from '@jupyterlab/services';
import { VDomModel } from '@jupyterlab/apputils';
import { Searcher, ISearchResult } from './query';
/**
 * Information about an extension
 */
export interface IEntry {
    name: string;
    description: string;
    python_package: string | null;
    installed: boolean;
    enabled: boolean;
    status: 'ok' | 'warning' | 'error' | 'deprecated' | null;
    latest_version: string;
    installed_version: string;
}
/**
 * Information about an installed extension
 */
export interface IInstalledEntry {
    name: string;
    description: string;
    enabled: boolean;
    core: boolean;
    latest_version: string;
    installed_version: string;
    status: 'ok' | 'warning' | 'error' | 'deprecated' | null;
}
/**
 * Extension actions that the server API accepts
 */
export declare type Action = 'install' | 'uninstall' | 'enable' | 'disable';
/**
 * Model for an extension list.
 */
export declare class ListModel extends VDomModel {
    constructor();
    /**
     * Utility function to check whether an entry can be updated.
     *
     * @param entry The entry to check.
     */
    static entryHasUpdate(entry: IEntry): boolean;
    /**
     * A readonly array of the installed extensions.
     */
    readonly installed: ReadonlyArray<IEntry>;
    /**
     * A readonly array containing the latest search result
     */
    readonly searchResult: ReadonlyArray<IEntry>;
    /**
     * Translate search results from an npm repository query into entries.
     *
     * @param res Promise to an npm query result.
     */
    protected translateSearchResult(res: Promise<ISearchResult>): Promise<{
        [key: string]: IEntry;
    }>;
    /**
     * Translate installed extensions information from the server into entries.
     *
     * @param res Promise to the server reply data.
     */
    protected translateInstalled(res: Promise<IInstalledEntry[]>): Promise<{
        [key: string]: IEntry;
    }>;
    /**
     * The current NPM repository search query.
     *
     * Setting its value triggers a new search.
     */
    query: string;
    /**
     * The current NPM repository search page.
     *
     * The npm repository search is paginated by the `pagination` attribute.
     * The `page` value selects which page is used.
     *
     * Setting its value triggers a new search.
     */
    page: number;
    /**
     * The NPM repository search pagination.
     *
     * The npm repository search is paginated by the `pagination` attribute.
     * The `page` value selects which page is used.
     *
     * Setting its value triggers a new search.
     */
    pagination: number;
    /**
     * The total number of results in the current search.
     */
    readonly totalEntries: number;
    /**
     * Make a request to the server for info about its installed extensions.
     */
    protected fetchInstalled(): Promise<IInstalledEntry[]>;
    /**
     * Initialize the model.
     */
    initialize(): void;
    /**
     * Update the current model.
     *
     * This will query the NPM repository, and the notebook server.
     *
     * Emits the `stateChanged` signal on succesfull completion.
     */
    protected update(): Promise<void>;
    /**
     * Send a request to the server to perform an action on an extension.
     *
     * @param action A valid action to perform.
     * @param entry The extension to perform the action on.
     */
    protected _performAction(action: string, entry: IEntry): Promise<IInstalledEntry[]>;
    /**
     * Install an extension.
     *
     * @param entry An entry indicating which extension to install.
     */
    install(entry: IEntry): void;
    /**
     * Uninstall an extension.
     *
     * @param entry An entry indicating which extension to uninstall.
     */
    uninstall(entry: IEntry): void;
    /**
     * Enable an extension.
     *
     * @param entry An entry indicating which extension to enable.
     */
    enable(entry: IEntry): void;
    /**
     * Disable an extension.
     *
     * @param entry An entry indicating which extension to disable.
     */
    disable(entry: IEntry): void;
    /**
     * Set to true if an error occurs when trying to query the NPM repository.
     */
    offline: boolean | undefined;
    /**
     * Contains an error message if an error occurred when updating the model.
     */
    errorMessage: string | undefined;
    /**
     * Whether the model has finished async initialization.
     */
    initialized: boolean;
    private _query;
    private _page;
    private _pagination;
    private _totalEntries;
    protected _installed: IEntry[];
    protected _searchResult: IEntry[];
    /**
     * Settings for connecting to the notebook server.
     */
    protected serverConnectionSettings: ServerConnection.ISettings;
    /**
     * A helper for performing searches of jupyterlab extensions on the NPM repository.
     */
    protected searcher: Searcher;
}
